import struct
import binascii
import psutil
import os
import threading
import time

import logging
logger = logging.getLogger('pythonagent.agent')

class Header(object):

    def __init__(self, context, apiReqId, awsReqId, funcName):

        self.apiReqId = apiReqId
        self.awsReqId = awsReqId
        self.funcName = funcName

        self.tags = "tierName={};ndAppServerHost={};appName={}".format(context.tier_name, context.server_name, funcName)

        self.agentType = 0  # Python
        self.messageType = None

        self.apiReqLen = None
        self.awsReqLen = None
        self.funcNameLen = None
        self.tagslength = None
        self.whLen = None

    def encode(self, msg_type="default"):

        # 3 Defaults added in arguments for testing wsgi start fp for havoc inbound delay

        if self.apiReqId is None:
            self.apiReqId = "default_apiReqId"
        if self.awsReqId is None:
            self.awsReqId = "default_awsReqId"
        if self.funcName is None:
            self.funcName = "default_funcName"

        # Message Type

        if msg_type == "default":
            self.messageType = 0
        elif msg_type == "havoc":
            self.messageType = 1
        elif msg_type == "init":
            self.messageType = 2
        elif msg_type == "heartbeat":
            self.messageType = 3
        else:
            raise Exception("Invalid msg_type in wrap header")

        self.apiReqLen = len(self.apiReqId)
        self.awsReqLen = len(self.awsReqId)
        self.funcNameLen = len(self.funcName)
        self.tagslength = len(self.tags)

        self.whLen = 20 + 2 + 2 + self.apiReqLen + self.awsReqLen + self.funcNameLen + self.tagslength + 1

        header_struct = struct.Struct('=ciiiiihh{}s{}s{}s{}s'.format(self.apiReqLen, self.awsReqLen,
                                                                     self.funcNameLen, self.tagslength))

        header_values = (bytes('^', "utf-8"),
                         self.whLen,
                         self.apiReqLen,
                         self.awsReqLen,
                         self.funcNameLen,
                         self.tagslength,
                         self.agentType,
                         self.messageType,
                         self.apiReqId.encode('utf-8'),
                         self.awsReqId.encode('utf-8'),
                         self.funcName.encode('utf-8'),
                         self.tags.encode('utf-8'))

        header_packet = header_struct.pack(*header_values)

        return header_packet


class StartTransactionMessage(object):

    def __init__(self, url):
        self.fp_header = "dummy_fp_header"
        self.url = url
        self.btHeaderValue = "dummy_btHeaderValue"
        self.ndCookieSet = ""
        self.nvCookieSet = ""
        self.correlationHeader = "dummy_correlationHeader"

    def encode(self):

        self.len_fp_header = len(self.fp_header)
        self.len_url = len(self.url)
        self.len_btHeaderValue = len(self.btHeaderValue)
        self.len_ndCookieSet = len(self.ndCookieSet)
        self.len_nvCookieSet = len(self.nvCookieSet)
        self.len_correlationHeader = len(self.correlationHeader)

        self.flowpathinstance = 0 # DUMMY
        self.qTimeMS = 0 # DUMMY
        self.startTimeFP = 0 # DUMMY

        # self.header_len = 48
        # self.total_len = self.header_len + self.len_url + 3

        self.header_len = 12
        self.total_len = 12 + 48 + self.len_fp_header + self.len_url + self.len_btHeaderValue + self.len_ndCookieSet + self.len_nvCookieSet + self.len_correlationHeader + 3

        self.msg_type = 2  # START TRANSACTION

        format_str = "=ciiiciiiiiiqqq{}s{}s{}s{}s{}s{}sc".format(self.len_fp_header, self.len_url, self.len_btHeaderValue, self.len_ndCookieSet, self.len_nvCookieSet, self.len_correlationHeader)

        start_transaction_message_struct = struct.Struct(format_str)

        start_transaction_message_values = (bytes('^', "utf-8"),
                                            self.header_len,
                                            self.total_len,
                                            self.msg_type,
                                            bytes('|', "utf-8"),
                                            self.len_fp_header,
                                            self.len_url,
                                            self.len_btHeaderValue,
                                            self.len_ndCookieSet,
                                            self.len_nvCookieSet,
                                            self.len_correlationHeader,
                                            self.flowpathinstance,
                                            self.qTimeMS,
                                            self.startTimeFP,
                                            self.fp_header.encode('utf-8'),
                                            self.url.encode('utf-8'),
                                            self.btHeaderValue.encode('utf-8'),
                                            self.ndCookieSet.encode('utf-8'),
                                            self.nvCookieSet.encode('utf-8'),
                                            self.correlationHeader.encode('utf-8'),
                                            bytes('\n', "utf-8"))

        start_transaction_message_packet = start_transaction_message_struct.pack(*start_transaction_message_values)

        return start_transaction_message_packet


class MethodEntryMessage(object):

    def __init__(self, methodName, query_string, url_parameter):
        self.methodName = methodName
        self.query_string = query_string
        self.urlParameter = url_parameter

    def encode(self):
        self.mid = 0  # DUMMY
        self.flowpathinstance = 0  # DUMMY
        self.threadId = 0 #threading.get_ident() # DUMMY
        self.startTime = 0  # DUMMY
        self.len_methodName = len(self.methodName)
        self.len_query_string = len(self.query_string)
        self.len_urlParameter = len(self.urlParameter)

        self.header_len = 12
        self.total_len = 12 + 40 + self.len_methodName + self.len_query_string + self.len_urlParameter + 3
        #self.total_len = 12 + 48 + self.len_methodName + self.len_query_string + self.len_urlParameter + 3


        self.msg_type = 0  # METHOD ENTRY
        # method_entry_message_struct = struct.Struct("=ciiiciqqqiii{}sc".format(self.len_methodName))

        
        format_str = "=ciiiciqqqiii{}s{}s{}sc".format(self.len_methodName, self.len_query_string, self.len_urlParameter)

        method_entry_message_struct = struct.Struct(format_str)

        method_entry_message_values = (bytes('^', "utf-8"),
                                       self.header_len,
                                       self.total_len,
                                       self.msg_type,
                                       bytes('|', "utf-8"),
                                       self.mid,
                                       self.flowpathinstance,
                                       self.threadId,
                                       self.startTime,
                                       self.len_methodName,
                                       self.len_query_string,
                                       self.len_urlParameter,
                                       self.methodName.encode('utf-8'),
                                       self.query_string.encode('utf-8'),
                                       self.urlParameter.encode('utf-8'),
                                       bytes('\n', "utf-8"))

        method_entry_message_packet = method_entry_message_struct.pack(*method_entry_message_values)

        return method_entry_message_packet


class MethodExitMessage(object):

    def __init__(self, methodName, backend_header, status, duration):

        self.methodName = methodName
        self.backend_header = backend_header
        self.requestNotificationPhase = "dummy_requestNotificationPhase"
        self.statusCode = status
        self.duration = duration
        self.mid = 0  # DUMMY
        self.eventType = 0  # DUMMY
        self.isCallout = 1  # DUMMY 1
        self.threadId = 0  # DUMMY threading.get_ident()
        self.flowpathinstance = 0  # DUMMY
        self.cpuTime = 0  # DUMMY

        self.len_methodName = None
        self.len_backend_header = None
        self.len_requestNotificationPhase = None

        self.tierCallOutSeqNum = 1  # DUMMY
        self.endTime = 0  # DUMMY

        self.header_len = None
        self.total_len = None

        self.msg_type = 1  # METHOD EXIT


    def encode(self):
        self.len_methodName = len(self.methodName)
        self.len_backend_header = len(self.backend_header)
        self.len_requestNotificationPhase = len(self.requestNotificationPhase)

        self.header_len = (4                                       # header_len(int)
                           + 4                                     # total_len(int)
                           + 4)                                    # msg_type(int)

        #self.total_len = 3 + 12 + 80 + self.len_methodName + self.len_backend_header + self.len_requestNotificationPhase

        self.total_len = (1                                        # ^ (1c)
                          + 4                                      # header_len(int)
                          + 4                                      # total_len(int)
                          + 4                                      # msg_type(int)
                          + 1                                      # | (1c)
                          + 4                                      # status_code(int)
                          + 4                                      # mid(int)
                          + 4                                      # eventType(int)
                          + 4                                      # isCallout(int)
                          + 8                                      # threadId(long)
                          + 8                                      # duration(long)
                          + 8                                      # flowpathinstance(long long)
                          + 8                                      # cpuTime(long long)
                          + 4                                      # len_methodName(int)
                          + 4                                      # len_backend_header(int)
                          + 4                                      # len_requestNotificationPhase(int)
                          + 8                                      # tierCallOutSeqNum(long long)
                          + 8                                      # endTime(long long)
                          + self.len_methodName                    # variable
                          + self.len_backend_header                # variable
                          + self.len_requestNotificationPhase      # variable
                          + 1)                                     # \n (1c)


        format_str = "=ciiiciiiiqqqqiiiqq{}s{}s{}sc".format(self.len_methodName, self.len_backend_header, self.len_requestNotificationPhase)

        method_exit_message_struct = struct.Struct(format_str)

        method_exit_message_values = (bytes('^', "utf-8"),
                                      self.header_len,
                                      self.total_len,
                                      self.msg_type,
                                      bytes('|', "utf-8"),
                                      self.statusCode,
                                      self.mid,
                                      self.eventType,
                                      self.isCallout,
                                      self.threadId,
                                      self.duration,
                                      self.flowpathinstance,
                                      self.cpuTime,
                                      self.len_methodName,
                                      self.len_backend_header,
                                      self.len_requestNotificationPhase,
                                      self.tierCallOutSeqNum,
                                      self.endTime,
                                      self.methodName.encode('utf-8'),
                                      self.backend_header.encode('utf-8'),
                                      self.requestNotificationPhase.encode('utf-8'),
                                      bytes('\n', "utf-8"))

        method_exit_message_packet = method_exit_message_struct.pack(*method_exit_message_values)
        logger.debug("status {} duration {}".format(self.statusCode, self.duration))

        return method_exit_message_packet


class EndTransactionMessage(object):

    def __init__(self, status_code):

        self.status_code = status_code
        self.flow_path_instance = 0  # DUMMY
        self.endTime = 0  # DUMMY
        self.cpuTime = 0  # DUMMY
        self.header_len = None
        self.total_len = None
        self.msg_type = 3  # END TRANSACTION

    def encode(self):

        self.header_len = (4                                       # header_len(int)
                           + 4                                     # total_len(int)
                           + 4)                                    # msg_type(int)

        #self.total_len = 3 + 12 + 32

        self.total_len = (1                                        # ^ (1c)
                          + 4                                      # header_len(int)
                          + 4                                      # total_len(int)
                          + 4                                      # msg_type(int)
                          + 1                                      # | (1c)
                          + 4                                      # status_code(int)
                          + 8                                      # flow_path_instance(long long)
                          + 8                                      # endTime(long long)
                          + 8                                      # cpuTime(long long)
                          + 1)                                     # \n (1c)

        end_transaction_message_struct = struct.Struct("=ciiiciqqqc")

        end_transaction_message_values = (bytes('^', "utf-8"),
                                          self.header_len,
                                          self.total_len,
                                          self.msg_type,
                                          bytes('|', "utf-8"),
                                          self.status_code,
                                          self.flow_path_instance,
                                          self.endTime,
                                          self.cpuTime,
                                          bytes('\n', "utf-8"))

        end_transaction_message_packet = end_transaction_message_struct.pack(*end_transaction_message_values)

        return end_transaction_message_packet


class TransactionEncodeHttpMessage(object):

    def __init__(self, bt, http_host, url):

        # bt = 140544547110928
        # http_host = "www.google.com"
        # url = "/"
        self.buffer = "dummy_buffer"
        self.type = "req"  # or "resp"


    def encode(self):
        self.statuscode = 0  # DUMMY
        self.buffer_len = len(self.buffer)
        self.type_len = len(self.type)
        self.flowpathinstance = 0  # DUMMY

        self.header_len = 12
        self.total_len = 12 + 20 + self.buffer_len + self.type_len + 3
        self.msg_type = 6  # HTTP CALLOUT BEGIN, 7 FOR END

        format_str = "=ciiiciiiq{}s{}sc".format(self.buffer_len, self.type_len)

        transaction_encode_http_message_struct = struct.Struct(format_str)

        transaction_encode_http_message_values = (bytes('^', "utf-8"),
                                                  self.header_len,
                                                  self.total_len,
                                                  self.msg_type,
                                                  bytes('|', "utf-8"),
                                                  self.statuscode,
                                                  self.buffer_len,
                                                  self.type_len,
                                                  self.flowpathinstance,
                                                  self.buffer.encode('utf-8'),
                                                  self.type.encode('utf-8'),
                                                  bytes('\n', "utf-8"))

        transaction_encode_http_message_packet = transaction_encode_http_message_struct.pack(*transaction_encode_http_message_values)


        return transaction_encode_http_message_packet


class HavocMessage(object):

    def __init__(self, havoc_header):
        self.havoc_header = havoc_header
        self.len_havoc_header = None

    def encode(self):

        self.len_havoc_header = len(self.havoc_header)

        format_str = "=c{}s".format(self.len_havoc_header)

        havoc_message_struct = struct.Struct(format_str)

        havoc_message_values = (bytes('^', "utf-8"),
                                self.havoc_header.encode('utf-8'))

        havoc_message_packet = havoc_message_struct.pack(*havoc_message_values)

        return havoc_message_packet


def create_start_transaction_message(context, bt_name, correlation_header):
    header = Header(context, context.api_request_id, context.aws_request_id, context.function_name)
    header_packet = header.encode()

    start_transaction_message = StartTransactionMessage(bt_name)
    start_transaction_message_packet = start_transaction_message.encode()

    full_packet = header_packet + start_transaction_message_packet

    return full_packet


def create_method_entry_message(context, bt, method, query_string, url_parameter):

    header = Header(context, context.api_request_id, context.aws_request_id, context.function_name)
    header_packet = header.encode()

    method_entry_message = MethodEntryMessage(method, query_string, url_parameter)
    method_entry_message_packet = method_entry_message.encode()

    full_packet = header_packet + method_entry_message_packet

    return full_packet


def create_method_exit_message(context, bt, method, backend_header, status, duration):
    header = Header(context, context.api_request_id, context.aws_request_id, context.function_name)
    header_packet = header.encode()

    method_exit_message = MethodExitMessage(method, backend_header, status, duration)
    method_exit_message_packet = method_exit_message.encode()

    full_packet = header_packet + method_exit_message_packet

    return full_packet


def create_end_transaction_message(context, bt, statuscode):
    header = Header(context, context.api_request_id, context.aws_request_id, context.function_name)
    header_packet = header.encode()

    start_transaction_message = EndTransactionMessage(statuscode)
    start_transaction_message_packet = start_transaction_message.encode()

    full_packet = header_packet + start_transaction_message_packet

    return full_packet


def create_havoc_message(context, havoc_header):

    header = Header(context, context.api_request_id, context.aws_request_id, context.function_name)
    msg_type = "havoc"
    header_packet = header.encode(msg_type)

    havoc_message = HavocMessage(havoc_header)
    havoc_message_packet = havoc_message.encode()

    full_packet = header_packet + havoc_message_packet

    return full_packet


def create_transaction_encode_http_message(context, bt, http_host, url):
    header = Header(context.api_request_id, context.aws_request_id, context.function_name)
    header_packet = header.encode()

    transaction_encode_http_message = TransactionEncodeHttpMessage(bt, http_host, url)
    transaction_encode_http_message_packet = transaction_encode_http_message.encode()

    full_packet = header_packet + transaction_encode_http_message_packet

    return full_packet


class InitMessage(object):

    def __init__(self):
        self.buffer = "dummy_buffer"

    def encode(self):
        #self.msg_type = "init"  # For informing proxy about agent status

        init_message_struct = struct.Struct("=c")
        init_message_values = (bytes('^', "utf-8"),)
        init_message_packet = init_message_struct.pack(*init_message_values)
        return init_message_packet

def create_init_message(context):
    header = Header(context, context.api_request_id, context.aws_request_id, context.function_name)
    logger.debug('function name inside wrap header {}'.format(context.function_name))
    msg_type = "init"
    header_packet = header.encode(msg_type)

    init_message = InitMessage()
    init_message_encode_packet = init_message.encode()

    full_packet = header_packet + init_message_encode_packet

    return full_packet


class HeartbeatMessage(object):

    def __init__(self):
        self.buffer = "dummy_buffer"

    def encode(self):

        heartbeat_message_struct = struct.Struct("=c")
        heartbeat_message_values = (bytes('^', "utf-8"),)
        heartbeat_message_packet = heartbeat_message_struct.pack(*heartbeat_message_values)
        return heartbeat_message_packet


def create_heartbeat_message(context):
    header = Header(context, context.api_request_id, context.aws_request_id, context.function_name)
    logger.debug('function name inside wrap header {}'.format(context.function_name))
    msg_type = "heartbeat"
    header_packet = header.encode(msg_type)

    heartbeat_message = HeartbeatMessage()
    heartbeat_message_encode_packet = heartbeat_message.encode()

    full_packet = header_packet + heartbeat_message_encode_packet

    return full_packet




"""Interceptor for httplib/http.client.

"""

from __future__ import unicode_literals
from ..base import ExitCallInterceptor
import time
#from functools import wraps
# from agent.internal.proxy import *


# import agent
class BotocoreClientInterceptor(ExitCallInterceptor):
    #print("INSIDE dynamodb.py BotocoreClientInterceptor class")

    def bind__create_api_method(py_operation_name, operation_name, service_model,
                                *args, **kwargs):
        return py_operation_name

    def _cav_boto_api_method(self, _create_api_method, ClientCreator,py_operation_name, operation_name, service_mode, *args, **kwargs):
        service_name = service_mode.service_name.lower()

        api_method = _create_api_method(ClientCreator, py_operation_name, operation_name, service_mode, *args, **kwargs)

        def _cav_wrapped_api_method(func, py_operation_name, _agent):
            # operation_name = py_operation_name
            agent = _agent
            def wrapper(*args, **kwargs):
                operation_name = py_operation_name
                #print("py_operation_name INSIDE wrapper :: ", operation_name)

                #print("LENGTH KWARGS INSIDE WRAPPER", len(kwargs))

                #for key in kwargs:
                #    print("%s = %s" % (key, kwargs[key]))

                if "TableName" in kwargs:
                    table_name = kwargs["TableName"]
                else:
                    table_name = "default"
                #print("operation_name, table_name INSIDE wrapper :: ", operation_name, table_name)
                #print("table_name INSIDE wrapper :: ", table_name)
                # print("INSIDE  _cav_wrapped_api_method :: ", (args))
                # print("LENGTH ARGS FOR _cav_wrapped_api_method::", args.items())
                # arg1 = args[0]

                start_time = None
                query_string = None
                method_name = None

                try:
                    if operation_name is None:
                        operation_name = "default"

                    method_name = "botocore.client." + str(operation_name)

                    query_string = str(table_name)+"."+str(operation_name)
                    start_time = time.time_ns()//1000000
                    self.agent.method_entry_http_callout(0, method_name, query_string, "dynamodb")

                except Exception as e:
                    print("Some error occurred inside wrapper in method entry call {}", e)

                results = func(*args, **kwargs)

                end_time = time.time_ns() // 1000000
                duration = end_time - start_time

                try:
                    from pythonagent.agent.probes.havoc.havoc_manager import NDNetHavocMonitor, NDHavocException
                    havoc_monitor = NDNetHavocMonitor.get_instance()
                    havoc_monitor.apply_outbound_service_failure(query_string)

                except Exception as e:
                    if isinstance(e, NDHavocException):
                        print("Havoc Exception {}".format(e))
                        zero_bt = 0
                        backend_header = "dynamodb"
                        status = 503

                        self.agent.method_exit_http_callout(zero_bt, method_name, backend_header, status,
                                                            duration)

                        self.agent.set_current_status_code(503)
                        raise e
                    else:
                        print("Non-Havoc Exception {}".format(e))

                try:
                    self.agent.method_exit_http_callout(0, method_name, "dynamodb", 200, duration)

                except Exception as e:
                    print("Some error occurred inside wrapper in method exit call {}", e)

                return results

            return wrapper

        wrapper = _cav_wrapped_api_method(api_method, py_operation_name, self.agent)

        if service_name == "dynamodb":
            return wrapper
        else:
            return api_method

def intercept_dynamodb(agent, mod):
    interceptor = BotocoreClientInterceptor(agent, mod.ClientCreator)
    interceptor.attach('_create_api_method', patched_method_name="_cav_boto_api_method")  



"""Interceptor for Redis.

"""

from __future__ import unicode_literals

from pythonagent.lang import str
from pythonagent.agent.probes.cache import CacheInterceptor
#from agent.internal.proxy import *
#import agent

class RedisConnectionInterceptor(CacheInterceptor):
    #self.agent.logger.info("going to get proxy instance (RedisConnectionInterceptor)")
    #proxy = Proxy.getInstance()
    #print("proxy handle = ", proxy)
    def __init__(self, agent, cls):
        super(RedisConnectionInterceptor, self).__init__(agent, cls, 'REDIS')

    def _send_packed_command(self, send_packed_command, connection, command,check_health=True): 
        #self.agent.logger.info('Modulename: RedisConnectionInterceptor class inside _send_packed_command function') 
        exit_call = None
        #backend= True
        bt = self.bt
        self.agent.logger.info("Modulename: RedisConnectionInterceptor class || bt value is {0}".format(bt)) 
        if not bt:
            bt = self.agent.get_current_bt()
        with self.log_exceptions():
            #self.agent.logger.info("inside _send_packed_command funciton")
            #self.agent.logger.info("send_packed_command: ", send_packed_command, "connection: ", connection, "command: ", command)
            self.agent.logger.info("Modulename: RedisConnectionInterceptor class || send_packed_command is {0} ".format(send_packed_command))
            self.agent.logger.info("Modulename: RedisConnectionInterceptor class || connection is {0}".format(connection))
            self.agent.logger.info("Modulename: RedisConnectionInterceptor class || connection.host is {0} ".format(connection.host))
            self.agent.logger.info("Modulename: RedisConnectionInterceptor class || connection.port is {0}".format(connection.port))
            self.agent.logger.info("Modulename: RedisConnectionInterceptor class || cmmand is {0} ".format(command))
            #self.agent.logger.debug("connection.host: ", connection.host, "connection.port: ", connection.port)
            self.agent.logger.debug("bt value: {0}".format(bt))
            if bt:
                #print("able to enter 1st if loop")
                try:
                    server_pool = ['%s:%s' % (connection.host, connection.port)]
                except AttributeError:
                    # For UnixDomainSocketConnection objects.
                    server_pool = [connection.path]
                #backend = self.get_backend(server_pool)
                backend = True
                if backend:
                    self.agent.logger.info("able to enter 2nd if loop")
                    #exit_call = self.proxy.db_call_begin(bt, connection.host,str(command))
                    exit_call = self.db_call_begin(bt, connection.host, str(command))
                    self.agent.logger.info("response after db_call_begin:{0} ".format(exit_call))
                    self.agent.logger.info("Modulename: RedisConnectionInterceptor class || exit_call is {0}".format(exit_call))

                    #exit_call = self.start_exit_call(bt, backend, operation=str(command))
        result = send_packed_command(connection, command,pythonagent_exit_call=exit_call)
        self.agent.logger.info("response after send_packed_command:".format(result))
        self.agent.logger.info("Modulename: RedisConnectionInterceptor class || result is ".format(result))
        #print("response before db_call_end:",exit_call = exit_call)
        #self.proxy.db_call_end(bt,exit_call)
        self.db_call_end(bt, exit_call)
        return result


def intercept_redis(agent, mod):
    RedisConnectionInterceptor(agent, mod.Connection).attach('send_packed_command')




from __future__ import unicode_literals
import contextlib
import sys
#import threading

#from lib import LazyWsgiRequest
#from appdynamics.agent.core.eum import inject_eum_metadata
#from appdynamics.agent.models.transactions import ENTRY_TORNADO
from ..base import EntryPointInterceptor
#from agent.internal.proxy import *
#import agent
#ENTRY_PYTHON_WEB = pb.PYTHON_WEB # need to ask ENTRY_PYTHON_WEB value
#ENTRY_TORNADO = ENTRY_PYTHON_WEB
try:
    import tornado.httputil
    import tornado.ioloop
   # import tornado.stack_context
   # import tornado.stack_context.StackContext
    import tornado.web
    import tornado.wsgi

    class TornadoFallbackHandlerInterceptor(EntryPointInterceptor):
        #proxy = Proxy.getInstance()
       # print("framework tornado_web: TornadoFallbackHandlerInterceptor function proxy instance inside  tornado_web.py : ", proxy)
       # print("framework tornado_web: TornadoFallbackHandlerInterceptor function proxy.lib inside class level variable inside tornado_web module : ", proxy.lib)
        # When using FallbackHandler, the RequestHandler's finish method is
        # never called.  Wrap the custom fallback callable to end the bt here.
        def _initialize(self, initialize, handler, fallback):
            self.agent.logger.info('Modulename: TornadoFallbackHandlerInterceptor class') 
            def _fallback(request):
                fallback(request)
                bt = self.bt
                self.agent.logger.debug("Modulename: TornadoFallbackHandlerInterceptor class || bt value is :{0}".format(bt))
                if bt:
                    #self.end_transaction(bt)
                    self.end_business_transaction(bt)
            initialize(handler, _fallback)

    class TornadoRequestHandlerInterceptor(EntryPointInterceptor):
        def __execute(self, _execute, handler, *args, **kwargs):
           # bt = self.proxy.start_business_transaction(ENTRY_TORNADO,
           #                             LazyWsgiRequest(tornado.wsgi.WSGIContainer.environ(handler.request)))
           # bt = self.start_business_transaction(handler.request,correlation_header=None)
            requests = handler.request
            bt = self.start_business_transaction(requests.path,'')
            #self.agent.logger.info("framework tornado_web: TornadoRequestHandlerInterceptor function print value of bt " .format(bt))
            self.agent.logger.debug("Modulename: TornadoRequestHandlerInterceptor class || bt value is :{0}".format(bt))

            #local_thread = threading.local()
            #local_thread.context = bt

            try:
                #print("framework tornado_web: TornadoRequestHandlerInterceptor  request in  ===>: ", handler.request)
                                                                                                                   
                #self.agent.logger.info("Modulename: TornadoRequestHandlerInterceptor class insinde __execute function request is :{0}".format(handler.request))
                # requests =handler.request
                #print("framework tornado_web: TornadoRequestHandlerInterceptor  request.PATH_INFO ===>: ", requests.path)
                #self.agent.logger.info("Modulename: TornadoRequestHandlerInterceptor class insinde __execute function request.PATH_INFO is :{0}".format(requests.path))
                #print("framework tornado_web: TornadoRequestHandlerInterceptor  request.QUERY_STRING ===>: ", requests.query_string)
                #self.agent.logger.info("Modulename: TornadoRequestHandlerInterceptor class insinde __execute function request.QUERY_STRING is :{0}".format(requests.query_string))
                # print("WSGI interceptor request.data ===>: ", request.data)
                #print("framework tornado_web: TornadoRequestHandlerInterceptor  request.host ===>: ", requests.host)
                self.agent.logger.info("Modulename: TornadoRequestHandlerInterceptor class insinde __execute function requests.host is :{0}".format(requests.host))
            except:
                pass

            if bt:
                @contextlib.contextmanager
                def current_bt_manager():
                    """Set and unset current_bt as tornado moves between execution contexts.

                    By wrapping the handler's execution with this we can ensure that whenever the
                    IOLoop is executing code for a particular BT, that BT is the 'current_bt'.
                    For more information see http://www.tornadoweb.org/en/stable/stack_context.html.

                    """
                    self.agent.set_current_bt(bt)
                    try:
                        yield
                    except:
                        # Currently can't figure out how to get here, so this code is untested.
                        #bt.add_exception(*sys.exc_info())
                        self.agent.logger.debug("Modulename: TornadoRequestHandlerInterceptor class || Exception is :{0}".format(*sys.exc_info()))
                        raise
                    finally:
                        self.agent.unset_current_bt()

               # with tornado.stack_context.StackContext(current_bt_manager):
                result = _execute(handler, *args, **kwargs)
            else:
                result = _execute(handler, *args, **kwargs)

            return result

        def _finish(self, finish, handler, *args, **kwargs):
            result = finish(handler, *args, **kwargs)
            bt = self.bt
            if bt:
                with self.log_exceptions():
                    self.handle_http_status_code(bt, handler._status_code, handler._reason)
                    #self.end_transaction(bt)
                    self.end_business_transaction(bt)
            return result

        def _flush(self, flush, handler, *args, **kwargs):
            with self.log_exceptions():
                if not handler._headers_written:
                    bt = self.bt
                    if bt:
                        headers = list(handler._headers.get_all())
                       # inject_eum_metadata(self.agent.eum_config, bt, headers)
                        handler._headers = tornado.httputil.HTTPHeaders(headers)
            return flush(handler, *args, **kwargs)

        def __handle_request_exception(self, _handle_request_exception, handler, e, *args, **kwargs):
            with self.log_exceptions():
                bt = self.bt
                if bt and not (hasattr(tornado.web, 'Finish') and isinstance(e, tornado.web.Finish)):
                    #bt.add_exception(*sys.exc_info())
                    self.agent.logger.debug("Modulename: TornadoRequestHandlerInterceptor class || Exception is :{0}".format(*sys.exc_info()))
            return _handle_request_exception(handler, e, *args, **kwargs)

    def intercept_tornado_web(agent, mod):
        #print("framework torando_web :  intercept_torando_web")
        TornadoRequestHandlerInterceptor(agent, mod.RequestHandler).attach(
            ['_execute', 'flush', '_handle_request_exception', 'finish'])
        TornadoFallbackHandlerInterceptor(agent, mod.FallbackHandler).attach('initialize')

except ImportError:
    def intercept_tornado_web(agent, mod):
        pass







"""Interceptors for SQL databases.

"""

from __future__ import unicode_literals

from . import mysql_connector, mysqldb, psycopg2, pymysql, dynamodb, botocores3

__all__ = ['pymysql','mysql_connector', 'mysqldb', 'psycopg2', 'dynamodb', 'botocores3']


from __future__ import unicode_literals
import functools

from pythonagent.lang import get_args, urlparse
from . import HTTPConnectionInterceptor
#from agent.internal.proxy import *
#import agent

try:
    import tornado.httpclient
   # import tornado.stack_context

    class AsyncHTTPClientInterceptor(HTTPConnectionInterceptor):
        #print("going to get  instance of AsyncHTTPClientInterceptor(HttpConnection Interceptor)")
        #self.agent.logger.debug('Modulename: AsyncHTTPClientInterceptor class') 
        #proxy = Proxy.getInstance()
        #print("proxy handle = ", proxy)
        def http_call_begin(self,url):
            #bt = self.bt
            bt =self.bt
           # bt =140138019344400
            self.agent.logger.info("Modulename: AsyncHTTPClientInterceptor class inside http_call_begin function bt is :{0}".format(bt))

            self.agent.logger.info("http tornado_http client module bt value .....{0} ".format(bt))
            #bt = self.agent.get_current_bt
            if not bt:
                return None

            parsed_url = urlparse(url)
            self.agent.logger.info("Modulename: AsyncHTTPClientInterceptor class inside http_call_begin function parsed_url is :{0}".format(parsed_url))

            port = parsed_url.port or ('443' if parsed_url.scheme == 'https' else '80')
            self.agent.logger.info("Modulename: AsyncHTTPClientInterceptor class inside http_call_begin function port is :{0}".format(port))
            #print("http tornado_httpclient module :AsyncHTTPClientInterceptor host",type(parsed_url.hostname))
            self.agent.logger.info("Modulename: AsyncHTTPClientInterceptor class inside http_call_begin function parsed_url.hostname is :{0}".format(parsed_url.hostname))
            #print("http tornado_httpclient module :AsyncHTTPClientInterceptor url",type( parsed_url.path))
            self.agent.logger.info("Modulename: AsyncHTTPClientInterceptor class inside http_call_begin function parsed_url.path is :{0}".format(parsed_url.path))
            #print("http tornado_httpclient module :AsyncHTTPClientInterceptor type of bt", type(bt))
            #backend = self.get_backend(parsed_url.hostname, port, parsed_url.scheme, url)
            #if not backend:
             #   return None

           # return super(AsyncHTTPClientInterceptor, self).http_call_begin(bt, parsed_url.hostname,url)
            return super().http_call_begin(bt,parsed_url.hostname,parsed_url.path)
        def http_call_end(self, exit_call, future):
            #super(AsyncHTTPClientInterceptor, self).http_call_end(exit_call, exc_info=future.exc_info())
            super().http_call_end(exit_call,exc_info=future.exc_info())
        def _fetch(self, fetch,client,request, raise_error=True, **kwargs):
            self.agent.logger.info('Modulename: AsyncHTTPClientInterceptor class inside _fetch function')
            exit_call = None
            with self.log_exceptions():
                is_request_object = isinstance(request, tornado.httpclient.HTTPRequest)
                url = request.url if is_request_object else request
                self.agent.logger.info("http tornado_httpclient module : _fetch method  url value {0}".format(url))
               # parsed_url = urlparse(url)         
                exit_call = self.http_call_begin(url)
                self.agent.logger.info("Modulename: AsyncHTTPClientInterceptor class inside _fetch function exit_call is :{0}".format(exit_call))

                self.agent.logger.info("http tornado_httpclient module : exit_call value ..........{0}".format(exit_call))
                if exit_call:
                    #self.agent.logger.info("http tornado_httpclient module : _fetch method  exit_call value", exit_call)
                    correlation_header = self.make_correlation_header(exit_call)
                    #print("http tornado_httpclient module : _fetch method  correlation_header value", correlation_header)
                    self.agent.logger.info("Modulename: AsyncHTTPClientInterceptor class inside _fetch function correlation_header is :{0}".format(correlation_header))
                    if correlation_header:
                        headers = request.headers if is_request_object else kwargs.setdefault('headers', {})
                        headers[correlation_header[0]] = correlation_header[1]
                        self.agent.logger.info("http tornado_httpclient module : _fetch method  headers[correlation_header[0]]".format(headers[correlation_header[0]]))

            # The `raise_error` kwarg was added in tornado 4.1.  Passing it by name on versions
            # prior to this cause it to be included in the `**kwargs` parameter to `fetch`.  This
            # dict is passed directly to the `HTTPRequest` constructor, which does not have
            # `raise_error` in its signature and thus raises a TypeError.
            if 'raise_error' in get_args(fetch):
                self.agent.logger.info("raise error.........".format(get_args(fetch)))
                future = fetch(client, request, raise_error=raise_error, **kwargs)
                #print(".............>",future)
                self.agent.logger.info('Modulename: AsyncHTTPClientInterceptor class inside _fetch function future is :'.format(future))
            else:
                future = fetch(client, request, **kwargs)
                #print("else future.......",future)
           # future._callbacks.insert(0, functools.partial(tornado.stack_context.wrap(self.http_call_end), exit_call))
            future._callbacks.insert(0, functools.partial(self.http_call_end, exit_call))
            return future

    def intercept_tornado_httpclient(agent, mod):
        # these methods don't normally return anything, but to be able to test that
        # the 'empty' interceptor defined below works properly, return a value here.
        #print("http torando_httpclient module : intercept_tornado_httpclient")
        return AsyncHTTPClientInterceptor(agent, mod.AsyncHTTPClient).attach('fetch', wrapper_func=None)
except ImportError:
    def intercept_tornado_httpclient(agent, mod):
        pass




import uuid
import os
import socket

import sys
import threading
import json
from .udp_message import create_start_transaction_message, create_end_transaction_message
from .udp_message import create_method_entry_message, create_method_exit_message
from .udp_message import create_havoc_message
from .udp_message import create_init_message
from .udp_message import create_heartbeat_message
import time
import logging

# from .udp_message import create_transaction_encode_http_message
udp_connection = None

logger = logging.getLogger('pythonagent.agent')


class UDPConnection(object):
    def __init__(self, agent_obj):

        if 'CAV_APP_AGENT_PROXYIP' in os.environ:
            self.cav_proxy_ip = os.environ['CAV_APP_AGENT_PROXYIP']
        else:
            self.cav_proxy_ip = "127.0.0.1"

        if 'CAV_APP_AGENT_PROXYPORT' in os.environ:
            self.cav_proxy_port = int(os.environ['CAV_APP_AGENT_PROXYPORT'])

        else:
            self.cav_proxy_port = 10000

        self.server_address_port = (self.cav_proxy_ip, self.cav_proxy_port)

        # Create a UDP socket at client side
        try:

            self.UDPClientSocket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
            logger.debug("Server Port {}".format(self.server_address_port))

            havoc_thread = threading.Thread(target=incoming_message_processor, name="havoc_thread",
                                            args=(self.UDPClientSocket,), daemon=True)
            havoc_thread.start()

            heartbeat_thread = threading.Thread(target=heartbeat_sender, name="heartbeat_thread",
                                                args=(self.UDPClientSocket, self.server_address_port, agent_obj),
                                                daemon=True)
            heartbeat_thread.start()


        except:
            logger.warning("Unable to create UDP Connection")

    def send(self, message, tx_type="non_start_fp"):

        # Send to server using created UDP socket
        try:
            self.UDPClientSocket.sendto(message, self.server_address_port)

        except:
            logger.warning("Unable to send UDP packet")


def heartbeat_sender(udp_socket, address_port, ag_obj):
    logger.info("heartbeat sender started")
    context = ag_obj.get_transaction_context()
    message = create_heartbeat_message(context)
    logger.info("heartbeat message {}".format(message))
    while True:
        try:
            logger.debug("heartbeat about to send")
            udp_socket.sendto(message, address_port)
            logger.info("heartbeat sent")
            time.sleep(10)
            logger.debug("heartbeat sleep completed")
        except Exception as e:
            logger.error("heartbeat exception", e)


def incoming_message_processor(udp_socket):
    logger.debug("New thread started")

    from pythonagent.agent.probes.havoc.havoc_manager import NDNetHavocMonitor  # Don't shift to top, it will cause circular import
    havoc_monitor = NDNetHavocMonitor.get_instance()
    # logger.debug("udp.py havoc id {}".format(id(havoc_monitor)))
    udp_socket.settimeout(2.0)
    buffer_size = 4096
    counter = 0

    while True:

        try:  # timeout (non-blocking behaviour) added because in warm-start, listener is not working sometimes
            counter += 1
            logger.debug("counter: {}".format(counter))
            logger.debug("Going to call recv from api for {} time".format(counter))
            logger.debug("Going to call recv from api")
            data, address = udp_socket.recvfrom(buffer_size)
            logger.debug("incoming received message = {} from address {}".format(data, address))
            logger.debug("after gettting message")

            output = str(data)

            if output == "b'Heart Beat Received'":
                logger.debug("ignoring heartbeat acknowledgement")
                continue
            try:

                output = output[2:-1]  # Remove b' from beginning and ' from end
                split_arr = output.split("\\n")

                header = split_arr[0]
                body = split_arr[1]

                body = body.replace("-", ":")

                header = header[:-1]  # Trim ; from end
                header_dict = {}

                header_split = header.split(";")

                count = 0
                for element in header_split:
                    if count == 0:
                        count += 1
                        continue  # Ignore first value: NetDiagnosticMessage2.0
                    else:
                        element_split = element.split(":")
                        key = element_split[0].strip()
                        value = element_split[1].strip()
                        header_dict[key] = value
                        count += 1
                havoc_monitor.parse_nethavoc_config(body, header_dict)

            except Exception as e:
                logger.warning("Unable to parse config: {}", e)
                logger.warning("unable to parse message except")

        except socket.timeout:
            logger.debug("recv timed out")  # This is not an exception, but a mechanism to implement timeout
            logger.warning("timed out after recvfrom")


def generate_bt():
    id = uuid.uuid4()
    id_int = id.int
    return id_int


def sdk_init(agent_obj):
    agent_obj.udp_connection = UDPConnection(agent_obj)
    logger.debug("called sdk init")
    context = agent_obj.get_transaction_context()
    logger.debug("transaction context dictionary inside sdk init in udp.py {}".format(context.function_name))
    message = create_init_message(context)
    logger.debug("first message to be sent to proxy {}".format(message))
    agent_obj.udp_connection.send(message)
    logger.debug("init sent")
    return udp_connection


def sdk_free(agent_obj):
    pass


def method_entry(agent_obj, bt, method, query_string, url_parameter):
    context = agent_obj.get_transaction_context()
    message = create_method_entry_message(context, bt, method, query_string, url_parameter)
    logger.info("method_entry {}".format(message))
    agent_obj.udp_connection.send(message)


def method_exit(agent_obj, bt, method, backend_header, status, duration):
    context = agent_obj.get_transaction_context()
    # message = create_method_exit_message(context, bt, method)
    message = create_method_exit_message(context, bt, method, backend_header, status, duration)
    logger.info("method_exit: status {}, duration {}, message {}".format(status, duration, message))
    agent_obj.udp_connection.send(message)


def start_business_transaction(agent_obj, bt_name, correlation_header):
    if os.environ['nd_init_done'] == '0':
        agent_obj.sdk_init()

    context = agent_obj.get_transaction_context()
    # message = udp_message.create_start_transaction_message(context, bt_name, correlation_header)
    message = create_start_transaction_message(context, bt_name, correlation_header)
    logger.info("start_business_transaction {}".format(message))
    agent_obj.udp_connection.send(message, "start_fp")

    bt = generate_bt()

    agent_obj.active_bts.add(bt)
    agent_obj.set_current_bt(bt)

    return bt


def end_business_transaction(agent_obj, bt):
    status_code = agent_obj.get_current_status_code()
    context = agent_obj.get_transaction_context()
    message = create_end_transaction_message(context, bt, status_code)
    agent_obj.logger.info("status_code {}, message {}".format(status_code, message))

    agent_obj.udp_connection.send(message)
    rc = 0  # DUMMY VALUE FOR SUCCESS
    agent_obj.reset_transaction_context()

    return rc


def store_business_transaction(agent_obj, bt, unique_bt_id):
    pass


def db_call_begin(agent_obj, bt, db_host, db_query):
    pass


def db_call_end(agent_obj, bt, ip_handle):
    pass


def http_call_begin(agent_obj, bt, http_host, url):
    handle = 1  # DUMMY NON ZERO VALUE
    return handle


def http_call_end(agent_obj, bt, ip_handle):
    pass


def havoc_message(agent_obj, havoc_header):
    context = agent_obj.get_transaction_context()
    message = create_havoc_message(context, havoc_header)
    logger.debug("havoc_message: {}".format(message))
    agent_obj.udp_connection.send(message)





from pythonagent.agent.probes import frameworks,http,sql,logging,mongodb,cache,Instrumentation,elasticdb

BT = (
    # Entry points
    ('flask', frameworks.intercept_flask),
    ('django.core.handlers.wsgi', frameworks.intercept_django_wsgi_handler),
    ('django.core.handlers.base', frameworks.intercept_django_base_handler),
    ('bottle', frameworks.intercept_bottle),
    ('tornado.web', frameworks.intercept_tornado_web),
    ('cherrypy', frameworks.intercept_cherrypy),    
    ('pyramid', frameworks.intercept_pyramid),
    ('falcon', frameworks.intercept_falcon),
    # HTTP exit calls
    ('httplib', http.intercept_httplib),
    ('http.client', http.intercept_httplib),
    ('urllib3', http.intercept_urllib3),
    ('requests', http.intercept_requests),
    ('tornado.httpclient', http.intercept_tornado_httpclient),
    #('boto.https_connection', http.intercept_boto),

    # Logging
    ('logging', logging.intercept_logging),

    # SQL exit calls
    ('psycopg2', sql.psycopg2.intercept_psycopg2_connection),
    ('pymysql.connections', sql.pymysql.intercept_pymysql_connections),
    ('mysql.connector.connection', sql.mysql_connector.intercept_mysql_connector_connection),
    ('MySQLdb.connections', sql.mysqldb.intercept_MySQLdb_connection),
    # MongoDB
    ('pymongo', mongodb.intercept_pymongo),
    ('elasticsearch',elasticdb.intercept_elastic),
    #Dynamo
    ('botocore.client', sql.dynamodb.intercept_dynamodb),
    ('botocore.endpoint', sql.botocores3.intercept_s3),

    #REDIS
    ('redis.connection', cache.intercept_redis),
)




"""Intercept boto to ensure that HTTPS is reported correctly.

"""

from __future__ import unicode_literals

from . import HTTPConnectionInterceptor


def intercept_boto(agent, mod):
    HTTPConnectionInterceptor.https_connection_classes.add(mod.CertValidatingHTTPSConnection)



from ..base import ExitCallInterceptor
from pythonagent.agent.probes.Instrumentation.parser import create_method_array
# from pythonagent.agent.probes.havoc.havoc_manager import NDNetHavocMonitor

import os
import configparser
import time

#configParser = configparser.RawConfigParser()
path = os.environ.get('ND_HOME')
path = path + '/config/ndsettings.conf'
print("path of  NDC config : ",path)
#configParser.read(path)
#path_for_insProfile = configParser.get('PYTHON_AGENT', 'path_profile')
myvars = {}

with open(path) as myfile:
    for line in myfile:
        name, var = line.partition("=")[::2]
        myvars[name.strip()] = var
try:
    #print("inside try")
    path_for_insProfile = myvars.get('path_profile').strip('\n')
except:
    #print("inside except")
    path_for_insProfile= os.environ.get('ND_HOME') + "/CavAgent/instrumentationprofile.json"
print('Profiler path at : ',path_for_insProfile)
#custom_methods_to_instrument = create_method_array('/home/cavisson/shop/my-shop/instrumentationprofile.txt')
custom_methods_to_instrument = create_method_array(path_for_insProfile)
print("ALL custom modules(Instrument)-> ", custom_methods_to_instrument)


class MethodInterceptor(ExitCallInterceptor):

    def _custom_method(self, original_method, *args, **kwargs):

        print("\n\n\n\nInside Custom Method\n\n\n")
        bt = self.bt

        if len(self.cls.__package__) >= 1:
            fqmforentry = str(self.cls.__package__) + "." + str(self.cls.__name__) + "." + str(original_method.__name__)
        else:
            #fqmforentry = "_." + str(self.cls.__name__) + "." + str(original_method.__name__)
            fqmforentry = "." + str(self.cls.__name__) + "." + str(original_method.__name__)

        # print("Inside custom method")

        before_time = time.time_ns() // 1000000

        # print("Before applying method call delay")
        # print("FQM: ", fqmforentry)

        from pythonagent.agent.probes.havoc.havoc_manager import NDNetHavocMonitor, NDHavocException
        havoc_monitor = NDNetHavocMonitor.get_instance()

        # havoc_monitor.apply_method_invocation_failure(fqmforentry)

        method = None
        if bt is not None:
            try:

                method = self.agent.method_entry(bt, fqmforentry)

                response = original_method(*args, **kwargs)

                havoc_monitor.apply_method_call_delay(fqmforentry)

            except Exception as e:
                self.agent.logger.exception('Modulename: MethodInterceptor class inside _custom_method got exception in orginal method')
                raise e  # Added by Samanvay on 06/04/2022 while debugging outbound havoc failure main -> http_callout -> request.get

            finally:

                after_time = time.time_ns() // 1000000
                duration = after_time - before_time
                print("Duration Calculated: ", duration)

                try:
                    havoc_monitor.apply_method_invocation_failure(fqmforentry)
                except Exception as e:
                    if isinstance(e, NDHavocException):
                        print("Havoc Exception {}".format(e))
                        status = 503
                        self.agent.method_exit(bt, fqmforentry, status, duration)
                        self.agent.set_current_status_code(503)
                        raise e
                    else:
                        print("Non-Havoc Exception {}".format(e))

                status = 0  # Needed for callouts only
                if self.agent.cav_env == "AWS_LAMBDA":
                    self.agent.method_exit(bt, fqmforentry, status, duration)
                if self.agent.cav_env == "NATIVE":
                    self.agent.method_exit(bt, fqmforentry)

                if method is not None:
                    if self.agent.cav_env == "AWS_LAMBDA":
                        self.agent.method_exit(bt, fqmforentry, status, duration)
                    if self.agent.cav_env == "NATIVE":
                        self.agent.method_exit(bt, fqmforentry)
        else:
            # self.agent.logger.info("did not find BT in ModuleInterceptor._custom_method")
            self.agent.logger.exception('Modulename: MethodInterceptor class inside _custom_method,,did not find BT in ModuleInterceptor._custom_method')
            response = original_method(*args, **kwargs)
        return response 


def intercept_instrumented_method(agent, mod):
    print("\n\n\n")
    print("intercept instrumented method")
    print("\n\n\n")

    for item in custom_methods_to_instrument:
        #print("CUSTOM METHOD TO INSTRUMENT ", item)

        #print("item name in custom_methods_to_intercept",item)
        #agent.logger.info("Modulename: MethodInterceptor class inside intercept_instrumented_method item name in custom_methods_to_intercept {0}".format(item))
        check = item.split(".")
        #print("array after split: ",check)
        if len(check) == 2:
            #print("list to inst 2", check)
            try:
                #print("ABOUT TO INTERCEPT INSTRUMENTED METHOD")
                MethodInterceptor(agent, mod).attach(check[1], patched_method_name='_custom_method')
            except:
                #agent.logger.debug("Instrumentation.__init__: class ".format(check[1])+" not present in module")
                agent.logger.exception('Modulename: MethodInterceptor class inside intercept_instrumented_method,Instrumentation.__init__: class ",check[1]," not present in module')

                continue    
               #MethodInterceptor(agent, mod).attach(check[1],wrapper_func='_custom_method')
        if len(check) == 3:
            #print("list to inst ", check)
            #MethodInterceptor(agent, getattr(mod, check[1])).attach(check[2],wrapper_func='_custom_method')
            try: 
                MethodInterceptor(agent, getattr(mod, check[1])).attach(check[2],patched_method_name='_custom_method')
            except:
                agent.logger.exception('Modulename: MethodInterceptor class inside intercept_instrumented_method, Instrumentation.__init__: method /class: ",check[1],".",check[2]," not present in module"')
                #agent.logger.debug("Instrumentation.__init__: method /class: ".format(check[1])+".".format(check[2])+" not present in module")
                continue





"""Interceptor for httplib/http.client.

"""

from __future__ import unicode_literals
from ..base import ExitCallInterceptor
import time
#from functools import wraps
# from agent.internal.proxy import *


# import agent
class BotocoreS3Interceptor(ExitCallInterceptor):
    print('inside botocore client interceptor class inside botos3')

    def _cav_s3_make_request(self, make_request, Endpoint, operation_model, request_dict):#, _agent):
        #agent = _agent
        print('operation model inside s3 make request', operation_model)
        print('operation model with extracted operation name', operation_model.name)
        method_name = 'botocore.endpoint' + str(operation_model)
        query_string = operation_model.name
        start_time = time.time_ns() // 1000000
        url = request_dict["url"]
        self.agent.method_entry_http_callout(0, method_name, query_string, url)
        endpiont_method = make_request(Endpoint, operation_model, request_dict)

        try:
            end_time = time.time_ns() // 1000000
            duration = end_time - start_time
            self.agent.method_exit_http_callout(0, method_name, "s3", 200, duration)

        except Exception as e:
            print("Some error occurred inside wrapper in method exit call {}", e)

        print('request_dict inside s3 make request', request_dict)
        return endpiont_method

def intercept_s3(agent, mod):
    print("DIR for mod DIR", mod)
    #interceptor = BotocoreClientInterceptor(agent, mod.ClientCreator)
    interceptor = BotocoreS3Interceptor(agent, mod.Endpoint)
    interceptor.attach('make_request', patched_method_name='_cav_s3_make_request')



"""Intercept the pymysql package.

"""

from __future__ import unicode_literals

from .dbapi import DbAPIConnectionInterceptor, DbAPICursorInterceptor


class PymysqlConnectionInterceptor(DbAPIConnectionInterceptor):
    def get_backend_properties(self, conn, *args, **kwargs):
        return conn.host, conn.port, conn.db, 'MYSQL'


class PymysqlCursorInterceptor(DbAPICursorInterceptor):
    def get_connection(self, cursor):
        return cursor.connection


def intercept_pymysql_connections(agent, mod):
    PymysqlConnectionInterceptor(agent, mod.Connection, PymysqlCursorInterceptor).attach('connect')



import json
import os
from datetime import datetime
import platform
import time
import threading 
import traceback
import sys
import multiprocessing
import psutil



#****creating a function for thread dump related variables ********

def dumpstacks():
 
    #vmName=my_system.system

    my_system = platform.uname()
    osName=my_system.system
    osVersion=my_system.version
    arch=my_system.machine
    b =multiprocessing.cpu_count()
    #l=psutil.getloadavg()
    l=os.getloadavg()
    res=sum(list(l))
    u=round(res,2)/3
    #ndbuild = sdk_getNdBuild()
    #ndbuild := C.GoString(C.sdk_getNdBuild())
    k=("Full thread dump Python Agent :- [ Dump Taken at : "+ datetime.now().strftime("%d/%m/%Y %H:%M:%S")+ " , Where BCIAgent build = \" 4.7.0 BUILD 93\" ")

    #k=("Full thread dump BCI Agent :- [ Dump Taken at : "+ datetime.now().strftime("%d/%m/%Y %H:%M:%S")+ " , Where BCIAgent build =\" " + ndbuild + " \""  )
    a=""

    nid="0x0"
    prio="5"
    vendor=""
    version=""   
   
    system_name = os.getenv('HOSTNAME')
 #//****append the thread details into list(code)*****//    
    code=[]
    o=[]
    al=[]
    for thread in threading.enumerate():
        
        """ if thread.daemon == True:
           s= 'wait'
        elif (thread.is_alive() == True ):
           s="Runnable"
        else:
           pass
        #if not thread.isDaemon():
           #s= 'wait'"""
        o.append(thread.name)

        if not thread.isDaemon():
           s= 'wait'
   
        elif thread.is_alive() == True:
           s='runable'
        
        elif thread.daemon == False:
           s='waiting'
        
        elif threading.Lock != null :
            s='locked'
       
        else:
           pass
        
    i=0
    for threadId, stack in sys._current_frames().items():
               
            code.extend(["\n\n'%s'" % o[i] + ' ' + "prio=%s"%prio + ' ' + "tid=%d"% threadId  + ' ' + "nid=%s"%nid + ' ' + "%s"%s+ ' ' + "\n" + "java.lang.Thread.State: %s "% s])
            i=i+1
               
            for filename, lineno, name, line in traceback.extract_stack(stack):
                 code.append('File: "%s", line %d, in %s, at %s' % (filename, lineno, name,datetime.now()))
                 if line:
                      code.append("  %s" % (line.strip()))
    print ("\n".join(code))

#//*****in dictionary thread dump entry page details comming ********//

    thread1_details = {"":k,"vmName":osName,"version":version ,"vendor":vendor , "osName":osName,"osVersion":osVersion,"arch":arch,"noOfProcessors":b,"SysLoadAvg":u,"deadlocked threads ":a }

#//*********all details dump into logs location************//

    def write_json(target_path, target_file, data):
        if not os.path.exists(target_path):
            try:
                os.makedirs(target_path)
            except Exception as e:
                print(e)
                raise
        with open(os.path.join(target_path, target_file), 'w') as f:
            json.dump(data, f)
            f.write("\n".join((code)))

    write_json('/opt/cavisson/netdiagnostics/logs/',
               datetime.now().strftime('thread_dump_%H_%M_%d_%m_%Y.txt'), thread1_details)

#dumpstacks()



import time
import os
import psutil
 
 
#def elapsed_since(start):
#    return time.strftime("%H:%M:%S", time.gmtime(time.time() - start))

def elapsed_since(start):
    #return time.strftime("%H:%M:%S", time.gmtime(time.time() - start))
    elapsed = time.time() - start
    if elapsed < 1:
        return str(round(elapsed*1000,2)) + " ms"
    if elapsed < 60:
        return str(round(elapsed, 2)) + " s"
    if elapsed < 3600:
        return str(round(elapsed/60, 2)) + " min"
    else:
        return str(round(elapsed / 3600, 2)) + " hrs"

def format_bytes(bytes):
    if abs(bytes) < 1000:
        return str(bytes)+" B"
    elif abs(bytes) < 1e6:
        return str(round(bytes/1e3,2)) + " kB"
    elif abs(bytes) < 1e9:
        return str(round(bytes / 1e6, 2)) + " MB"
    else:
        return str(round(bytes / 1e9, 2)) + " GB"
 
def get_process_memory():
    process = psutil.Process(os.getpid())
    mem_info = process.memory_info()
    return mem_info.rss
 
 
def profile(func):
    def wrapper(*args, **kwargs):
        mem_before = get_process_memory()
        start = time.time()
        result = func(*args, **kwargs)
        elapsed_time = elapsed_since(start)
        mem_after = get_process_memory()
        print("{0}: memory before: {1}, after: {2}, consumed: {3}; exec time: {4}".format(
            func.__name__,
            format_bytes(mem_before), format_bytes(mem_after), format_bytes(mem_after - mem_before),
            elapsed_time))
        return result
    return wrapper





"""Intercept the mysql-connector-python package.

"""

from __future__ import unicode_literals

from .dbapi import DbAPIConnectionInterceptor, DbAPICursorInterceptor


class MySQLConnectionInterceptor(DbAPIConnectionInterceptor):
    def get_backend_properties(self, conn, *args, **kwargs):
        return conn._host, conn._port, conn._database, 'MYSQL'


class MySQLCursorInterceptor(DbAPICursorInterceptor):
    def get_connection(self, cursor):
        return cursor._connection


def intercept_mysql_connector_connection(agent, mod):
    MySQLConnectionInterceptor(agent, mod.MySQLConnection, MySQLCursorInterceptor).attach('_open_connection')



"""Interceptor for Cherry framework.

"""
from __future__ import unicode_literals
import sys

from pythonagent.agent.probes.frameworks.wsgi import WSGIInterceptor
from pythonagent.agent.probes.base import BaseInterceptor

class CherrypyExceptionAdder(BaseInterceptor):
    def add_exception(self, func, *args, **kwargs):
        with self.log_exceptions():
            bt = self.bt
            self.agent.logger.info("Modulename: cherry exception  class || bt  value is {0} :".format(bt))
            if bt:
                bt.add_exception(*sys.exc_info())
                self.agent.logger.debug("Exception occured as  !!! {0}".format(sys.exc_info()))
        return func(*args, **kwargs)


def intercept_cherrypy(agent, mod):
    WSGIInterceptor(agent, mod.Application).attach('__call__')
    CherrypyExceptionAdder(agent, mod._cprequest.Request).attach('handle_error', patched_method_name='add_exception')
    CherrypyExceptionAdder(agent, mod.HTTPError).attach('set_response', patched_method_name='add_exception')



import os
import ast
import json

import sys
py_ver = sys.version_info[0]

local = False

def find(path):
    modules = []
    default_path = []
    for root, dirs, files in os.walk(path):
        for file in files:
            # print(file)
            if file.endswith(".py"):
                default_path.append(os.path.join(root, file))
                file = file[:-3]  # Remove .py extension
                modules.append(file)

    # print("modules", modules)
    # print("\n\n")
    # print("default_path", default_path)
    # print("\n\n")

    path_module_dict = dict(zip(default_path, modules))

    data = {
        "modules": [
            {
                "moduleName": path_module_dict[filename],
                "moduleLevelMethod": [
                    {
                        "name": f_name.name,
                        "signature": [a.arg if py_ver == 3 else a.id for a in f_name.args.args]
                        # "signature": [[a.arg for a in f_name.args.args],[f_name.returns]]

                    }
                    for f_name in [n for n in ast.parse(open(filename).read()).body if isinstance(n, ast.FunctionDef)]
                ],
                "classes": [
                    {
                        "className": cname.name,
                        "classLevelMethods": [
                            {
                                "name": clm_name.name,
                                "signature": [a.arg if py_ver == 3 else a.id for a in clm_name.args.args]
                                # "signature": [[a.arg for a in clm_name.args.args], [clm_name.returns]]

                            }
                            for clm_name in [n for n in cname.body if isinstance(n, ast.FunctionDef)]
                        ]
                    }
                    for cname in [n for n in ast.parse(open(filename).read()).body if isinstance(n, ast.ClassDef)]
                ]
            }
            for filename in default_path
        ]
    }

    json_object = json.dumps(data, indent=2)
    # print("\n\n\n\n")
    # print(json_object)
    # return str(json_object)

    with open(os.environ.get('ND_HOME') + "/CavAgent/instrumentationprofile.json", "w") as outfile:
        outfile.write(json_object)
        #print(outfile)

if local:
    path = "/home/cavisson/shop/my-shop/pythonagent/"
    find(path)



"""Interceptor for Django.

"""


from __future__ import unicode_literals
import sys

from pythonagent.agent.probes.frameworks.wsgi import WSGIInterceptor
from pythonagent.agent.probes.base import BaseInterceptor
import logging
#from django.conf import settings



def add_exception(interceptor, exc_info):
    with interceptor.log_exceptions():
        bt = interceptor.bt
        if bt:
            bt.add_exception(*exc_info)

#from django.utils.deprecation import MiddlewareMixin






class DjangoBaseHandlerInterceptor(BaseInterceptor):
    def _load_middleware(self, load_middleware, base_handler):
        self.agent.logger.info("Loading middleware of django !")
        self.agent.logger.info('Modulename: DjangoBaseHandlerInterceptor class inside _load_middleware function : loading middleware of django !')
        #settings.MIDDLEWARE.append('agent.my_middleware.CustomMiddleware')
        load_middleware(base_handler)

        base_handler._exception_middleware.insert(0, pythonagentDjangoMiddleware(self).process_exception)
        #base_handler._request_middleware.insert(0, pythonagentDjangoMiddleware(self).process_request)
        #base_handler._response_middleware.insert(0, pythonagentDjangoMiddleware(self).process_response)
        #base_handler._template_response_middleware.insert(0, pythonagentDjangoMiddleware(self).process_template_response)
        #base_handler._view_middleware.insert(0, pythonagentDjangoMiddleware(self).process_view)



    def _handle_uncaught_exception(self, handle_uncaught_exception, base_handler, request, resolver, exc_info):
        add_exception(self, exc_info)
        return handle_uncaught_exception(base_handler, request, resolver, exc_info)


class DjangoExceptionInterceptor(BaseInterceptor):
    def _handle_uncaught_exception(self, handle_uncaught_exception, request, resolver, exc_info):
        add_exception(self, exc_info)
        return handle_uncaught_exception(request, resolver, exc_info)




class pythonagentDjangoMiddleware(object):
    def __init__(self, interceptor):
        self.interceptor = interceptor
        self.logger = logging.getLogger('pythonagent.agent')

    def process_exception(self, request, exception):
        add_exception(self.interceptor, sys.exc_info())
        #print("Django application process_request called and request as= ", request)


def intercept_django_wsgi_handler(agent, mod):
    WSGIInterceptor(agent, mod.WSGIHandler).attach('__call__')



def intercept_django_base_handler(agent, mod):
    base_handler_methods = ['load_middleware']

    try:
        import django.core.handlers.exception
        if hasattr(django.core.handlers.exception, 'handle_uncaught_exception'):
            DjangoExceptionInterceptor(agent, django.core.handlers.exception).attach('handle_uncaught_exception')
        else:
            base_handler_methods.append('handle_uncaught_exception')
    except ImportError:
        base_handler_methods.append('handle_uncaught_exception')

    DjangoBaseHandlerInterceptor(agent, mod.BaseHandler).attach(base_handler_methods)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pythonagent", # Replace with your own username
    version="0.0.15",
    author="Cavisson System",
    author_email="pythonagent@cavisson.com",
    description="A small example package",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/pypa/sampleproject",
    classifiers=[
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    packages=setuptools.find_packages(),
    #package_data={'pythonagent': ['*.txt']},
    include_package_data=True,
    #python_requires='>=3.6',
    python_requires='>=2.6, !=3.0.*, !=3.1.*, !=3.2.*',
)



import os
import ast
import json
import ctypes
import sys
import time
import pyclbr
import psutil
import datetime
import hashlib

def newmodulefinder():
    import sys
    from modulefinder import ModuleFinder
    finder = ModuleFinder()
    finder.run_script(sys.argv[0])
    a = finder.modules.items()
    modulelist = []
    #print('Loaded modules:')
    for name, mod in finder.modules.items():
        if mod.__file__ is None:
            pass
        else:
            modulelist.append(mod.__file__)
    return modulelist


def find():
    p = psutil.Process(os.getpid())
    procname = p.name()
    pathforfile =  "/opt/cavisson/netdiagnostics/logs/"
    filename = str(procname+"_AD.txt")
    file_to_open = os.path.join(pathforfile, filename)
    f = open(file_to_open, 'w')
    f.writelines("************************************  AUTO-DISCOVERY  **************************************** \n")
    nylist = newmodulefinder()
    for path in nylist:
        module_path = os.path.split(os.path.abspath(path))[0]
        modulefullname = os.path.split(os.path.abspath(path))[1]
        module_name = modulefullname.split('.')[0]
        modleres = pyclbr.readmodule(module_name, path=[module_path])
        #f.writelines("************************************  AUTO-DISCOVERY  **************************************** \n")
        if len(modleres)>0:
            import sys
            sys.path
            for i in modleres.items():
                aa = modleres[str(i[0])].methods.keys()
                if len(aa)>0:
                    joined_string = ",".join(aa)
                    outputfromad = module_name+"|"+i[0]+"|"+joined_string
                    f.writelines(f"{outputfromad}  \n")
                    #print(outputfromad)
                else :
                    pass
        else:
            pass
        
        modleres_ex = pyclbr.readmodule_ex(module_name, path=[module_path])
        li =[]
        for k, v in modleres_ex.items():
            if str(v) != None and "pyclbr.Function" in str(v):
                li.append(k)
            else:
                pass
        joined_stringf = ",".join(li)
                #joined_stringf = ",".join(li)
        outputfromadf = module_name+"|"+joined_stringf
                #outputfromadf = module_name+"|"+k
        f.writelines(f"{outputfromadf} \n")

    return outputfromad




"""Entry-point interceptors for various web/application frameworks.

"""

from __future__ import unicode_literals

from .django import intercept_django_wsgi_handler, intercept_django_base_handler
from .flask import intercept_flask
from .bottle import intercept_bottle
from .tornado_web import intercept_tornado_web
from .cherry import intercept_cherrypy
from .pyramid import intercept_pyramid
from .falcon import intercept_falcon
#__all__ = [
#    'intercept_flask',
#    'intercept_django_wsgi_handler',
#    'intercept_django_base_handler',
#    'intercept_bottle',
#    'intercept_tornado_web',
#]



import requests
import boto3
from flask import Flask


def http_callout():
    print("Client Lambda - HTTP Callout Begin")
    x = requests.get('https://www.facebook.com/')
    # x = requests.put('https://httpbin.org/put', data={'key': 'value'})
    # x = requests.post('https://www.w3schools.com/python/demopage.php', data={'key': 'value'})
    # x = requests.delete('https://httpbin.org/delete', data={'key': 'value'})
    print("Client Lambda - HTTP Callout End")


def first():
    print("Client Lambda - First Function")


def second():
    print("Client Lambda - Second Function")


def db_callout():
    print("Client Lambda - DB Callout Begin")
    client = boto3.client('dynamodb')
    print("client METHODS FOR THIS SESSION::", dir(client))
    data = client.scan(TableName='Person')
    item = {
        "id": {"N": "556"},
        "address": {"S": "United States"},
        "firstName": {"S": "John"},
        "lastName": {"S": "Doe"},
        "age": {"N": "30"}
    }
    response = client.put_item(TableName='Person', Item=item)
    print("Client Lambda - DB Call out Begin")


def flask_wsgi():
    app = Flask(__name__)

    @app.route("/")
    def hello_world():
        return "<p>Hello, World!</p>"

    app.run(port=6063)


"""Base interceptor for distributed caches (typically, key-value stores).

"""

from __future__ import unicode_literals

#from appdynamics.agent.models.exitcalls import EXIT_CACHE, EXIT_SUBTYPE_CACHE
from ..base import ExitCallInterceptor

EXIT_CACHE= 1
EXIT_SUBTYPE_CACHE ='CACHE'
class CacheInterceptor(ExitCallInterceptor):
    """Base class for cache interceptors.

    Extra Parameters
    -----------------
    vendor : string
        The vendor name of this cache backend e.g. MEMCACHED.

    """

    backend_name_format_string = '{SERVER POOL} - {VENDOR}'

    def __init__(self, agent, cls, vendor):
        self.vendor = vendor
        super(CacheInterceptor, self).__init__(agent, cls)

    def get_backend(self, server_pool):
        """

        Parameters
        ----------
        server_pool : list of str

        """
        self.agent.logger.info('Modulenam CacheInterceptor class inside get_backend') 
        backend_properties = {
            'VENDOR': self.vendor,
            'SERVER POOL': '\n'.join(server_pool),
        }
        self.agent.logger.info("Modulenam CacheInterceptor class inside get_backend backend_properties is {0}".format(backend_properties))
        return self.agent.backend_registry.get_backend(EXIT_CACHE, EXIT_SUBTYPE_CACHE, backend_properties,self.backend_name_format_string)
        # return None


from .redis import intercept_redis


__all__ = ['intercept_redis']




import pythonagent.agent as agent
_agent = agent.get_agent_instance()


def methodwrapper(func):
    print(func.__module__)
    fqm = str(func.__module__) +"."+ str(func.__name__)
    bt = _agent.get_current_bt()
    #bt = "000001"
    def inner(*args, **kwargs):
        try:
            _agent.method_entry(bt, fqm)
        except Exception as e:
            print(e)
        rc = func(*args, **kwargs)
        try:
            _agent.method_exit(bt, fqm, 200)
        except Exception as e:
            print(e)
        return rc
        
    return inner



from __future__ import unicode_literals
import imp
import os
import warnings
import logging
import sys
from subprocess import call
import subprocess
import uuid
import datetime
import time

is_lambda = True

if is_lambda:
    os.environ.setdefault("ND_HOME", "/opt/python/pythonagent/cavisson/netdiagnostics")
    sys.path.insert(1, '/opt')

os.environ.setdefault("CAVISSON_APP_NAME", os.getenv("AWS_LAMBDA_FUNCTION_NAME", ""))
os.environ.setdefault("CAVISSON_NO_CONFIG_FILE", "true")
os.environ.setdefault("CAVISSON_DISTRIBUTED_TRACING_ENABLED", "true")
os.environ.setdefault("CAVISSON_SERVERLESS_MODE_ENABLED", "true")
os.environ.setdefault("CAVISSON_TRUSTED_ACCOUNT_KEY", os.getenv("CAVISSON_ACCOUNT_ID", ""))

_agent = None
logger = None

try:
    import pythonagent.agent as agent
    agent.configure()
    #_agent = agent.bootstrap()
    _agent = agent.get_agent_instance()
    #_agent.init_sent = True

except:
    print("import error in Agent")
    logger = logging.getLogger('pythonagent.agent')
    logger.exception('Exception in agent startup.')
finally:
    pass

if _agent is None:
    raise ValueError("Failed to initialize agent")

from pythonagent.agent.internal.agent import TransactionContext

# from pythonagent.agent.probes.havoc.havoc_manager import NDNetHavocMonitor, NDHavocException


def get_handler():
    print("inside get_handler function first", sys.executable)
    modandpackage = os.path.abspath('.')
    mod = sys.modules[__name__]
    print("inside get_handler function mod", mod)
    try:
        sys.path.remove(os.path.dirname(__file__))
    except ValueError:  # directory not in sys.path
        pass


    if (
            "CAVISSON_LAMBDA_HANDLER" not in os.environ
            or not os.environ["CAVISSON_LAMBDA_HANDLER"]
    ):
        raise ValueError(
            "No value specified in CAVISSON_LAMBDA_HANDLER environment variable"
        )

    try:
        module_path, handler_name = os.environ["CAVISSON_LAMBDA_HANDLER"].rsplit(
            ".", 1
        )
        print("inside get_handler function module_path, handler name: ", module_path, handler_name)

    except ValueError:
        raise ValueError(
            "Improperly formated handler value: %s"
            % os.environ["CAVISSON_LAMBDA_HANDLER"]
        )

    file_handle, pathname, desc = None, None, None

    try:
        for segment in module_path.split("."):
            if pathname is not None:
                pathname = [pathname]

            file_handle, pathname, desc = imp.find_module(segment, pathname)

        if file_handle is None:
            module_type = desc[2]
            if module_type == imp.C_BUILTIN:
                raise ImportError(
                    "Cannot use built-in module %s as a handler module" % module_path
                )

        print("inside get_handler function module_path, file_handle, pathname, desc", module_path, file_handle, pathname, desc)

        module = imp.load_module(module_path, file_handle, pathname, desc)
        print("inside get_handler function module in cavagent_lambda_wrapper ", module)
        

    except Exception as e:
        print("Module not found")
        raise ImportError("Failed to import module '%s': %s" % (module_path, e))
    finally:
        if file_handle is not None:
            file_handle.close()

    try:
        handler = getattr(module, handler_name)
        print("inside get_handler function handler", handler)
    except AttributeError:
        print("NO HANDLER FOUND")
        raise AttributeError(
            "No handler '%s' in module '%s'" % (handler_name, module_path)
        )

    return handler, handler_name, modandpackage


def local_customer_application(event, context):
    from .run import db_callout, flask_wsgi, http_callout, first
    print("inside local_customer_application function Lambda Handler Function")

    first()
    # second()
    # http_callout()
    # db_callout()
    # flask_wsgi()

    return "Nothing"


def local_lambda_handler():
    #for i in range(100):
    #    print(1)
    class TestObj:
        def __init__(self):
            self.__module__ = "test"

    t = TestObj()
    return t , "test", "test"

# Greedily load the handler during cold start, so we don't pay for it on first invoke


def wrapped_handler(event, context, is_lambda):
    if is_lambda:
        a, b, c = get_handler()
    else:
        a, b, c = local_lambda_handler()
    return a, b, c


def handler(event, context):
    context_obj = TransactionContext()
    start_time = datetime.datetime.now()

    if context.aws_request_id is not None:
        context_obj.aws_request_id = context.aws_request_id
    else:
        context_obj.aws_request_id = "default_aws_request_id"

    if context.function_name is not None:
        context_obj.function_name = context.function_name
    else:
        context_obj.function_name = "default"
    print("function name inside context object in cavwrapper handler", context_obj.function_name)


    try:
        # context_obj.function_name = event['requestContext']['path']
        context_obj.url_path = event['requestContext']['path']

    except:
        if "path" in event:
            context_obj.url_path = event["path"]
        else:
            context_obj.url_path = context.function_name

    if context.log_stream_name is not None:
        context_obj.api_request_id = context.log_stream_name
    else:
        new_id = uuid.uuid4()
        id_int = new_id.int
        context_obj.api_request_id = id_int

    _agent.set_transaction_context(context_obj)

    from pythonagent.agent.probes.havoc.havoc_manager import NDNetHavocMonitor, NDHavocException
    havoc_monitor = NDNetHavocMonitor.get_instance()

    bt = None

    try:
        bt = _agent.start_business_transaction(context_obj.url_path, "")
    except Exception as e:
        print("Error occurred in Start Business transaction Call {}".format(e))

    # import time
    # print("Start fp done, sleep 30")
    # time.sleep(30)

    try:
        handle, handler_name, modandpackage = wrapped_handler(event, context, is_lambda)
    except Exception as e:
        print("Error occurred in Calling Wrapped Handler {}".format(e))


    print("inside handler function handle, handler_name, modandpackage IN CAVAGENT_LAMBDA_WRAPPER", handle, handler_name, modandpackage)
    #fqmmethodentry = str(modandpackage) + "." + str(handler_name)
    fqmmethodentry = str(handle.__module__) + "." + str(handler_name)

    _agent.method_entry(bt, fqmmethodentry)

    if is_lambda:
        try:
            response = handle(event, context)  # Client Lambda Function
            print("inside handler function RESPONSE IN LAMBDA IN CAVAGENT_LAMBDA_WRAPPER", response)
        except Exception as e:
            print("Error while invoking handler {}".format(e))
            _agent.method_exit(bt, fqmmethodentry, 503)
            _agent.set_current_status_code(503)
            rc = _agent.end_business_transaction(bt)
            havoc_monitor.refresh_configs(_agent)
            raise e
    else:
        response = local_customer_application(event, context)
        print("inside handler function RESPONSE IN LAMBDA IN CAVAGENT_LAMBDA_WRAPPER", response)

    status_code = 200
    try:
        if "statusCode" in response:
            status_code = response["statusCode"]

        print("Inside handler function: response inside get_handler function firststatus code ", status_code)

    except Exception as e:
        print("Error occurred in get status code {}".format(e))



    ###############################
    # HAVOC
    ###############################

    # Apply Inbound Delay --> agent/start_business_transaction

    # Apply Inbound Failure

    try:
        print("Url request: ", context_obj.url_path)
        havoc_monitor.apply_inbound_service_failure(context_obj.url_path)

    except Exception as e:
        if isinstance(e, NDHavocException):
            print("Havoc Exception {}".format(e))
            _agent.method_exit(bt, fqmmethodentry, 503)
            _agent.set_current_status_code(503)
            rc = _agent.end_business_transaction(bt)
            havoc_monitor.refresh_configs(_agent)
            raise e
        else:
            print("Non-Havoc Exception {}".format(e))

    # Apply Outbound Delay --> agent/method_entry_http_callout

    # Apply Outbound Failure --> agent/probes/base.py/http_call_end
    #                        --> agent/probes/sql/dynamodb.py/_cav_boto_api_method

    # Apply Method Delay --> Instrumentation/__init__

    # Apply Method Failure --> Instrumentation/__init__

    ####################################################################################

    # havoc_monitor.apply_outbound_service_failure("sample_hostname")

    havoc_monitor.refresh_configs(_agent)

    try:
        _agent.method_exit(bt, fqmmethodentry, status_code)
        _agent.set_current_status_code(status_code)
        rc = _agent.end_business_transaction(bt)

    except Exception as e:
        print("Error occurred in End Business transaction Call {}".format(e))

    return response





HELP_OPTIONS = ('--help', '-h', '-?')

items = lambda d: d.items()
keys = lambda d: d.keys()

class CommandInvocationError(Exception):
    pass


class CommandExecutionError(Exception):
    pass


def parse_options(option_descrs, args):
    options = {}

    required = set()
    short_flags = {}
    for opt, opt_descr in items(option_descrs):
        if isinstance(opt_descr, dict):
            if 'short' in opt_descr:
            # short_flags[opt_descr['short'][0]] = opt
                short_flags[opt_descr['short']] = opt
            if opt_descr.get('required', False):
                required.add(opt)

    i = 0
    argc = len(args)

    while i < argc and args[i][:1] == '-':  # While there's an arg and it's an option...
        arg = args[i]

        if arg == '--' or arg == '-':       # The option to end all options.
            i += 1                          # Skip the dash.
            break

        if arg in HELP_OPTIONS:             # Help options short circuit everything.
            return {'help': True}, []

        if arg[1] == '-':                   # Long option.
            opt = arg[2:]
        else:                               # Short option.
            #             opt = short_flags.get(arg[1:])
            opt = short_flags.get(arg[1:])

        if opt not in option_descrs:
            raise CommandInvocationError('unrecognized option: %s' % arg)

        option_descr = option_descrs[opt]

        if isinstance(option_descr, dict) and option_descr.get('value'):
            value_label = option_descr.get('value_help', '<value>')
            i += 1

            if i >= argc:
                raise CommandInvocationError('missing value for option: %s %s' % (arg, value_label))

            options[opt] = args[i]
        else:
            options[opt] = True

        i += 1

    missing = ['--%s' % r for r in required - set(keys(options))]

    if missing:
        plural = 's' if len(missing) != 1 else ''
        missing = ', '.join(missing)
        raise CommandInvocationError('missing required option%s: %s' % (plural, missing))

    return options, args[i:]










"""Intercept the Python logging module.

"""

import logging

from ..base import BaseInterceptor


class LoggingInterceptor(BaseInterceptor):
    LOGGING_ERROR_DISPLAY_NAME = '{logger_name} [{level_name}]'

    def _callHandlers(self, callHandlers, logger, record):
        with self.log_exceptions():
            if not record.name.startswith('pythonagent') and self.bt:
                self.agent.logger.info("error message as {}".format(record.getMessage()))
                self.add_logged_error(record)

        try:
            if type(record.msg) != str:
                record.msg = str(record.msg)

            context = self.agent.get_transaction_context()
            recordReqIdmsg = f" [ RequestId : {context.aws_request_id} ] " + record.msg
            record.msg = recordReqIdmsg
        except Exception as e:
            print(e)

        callHandlers(logger, record)

    def add_logged_error(self, record):
        """Add an error to the BT if the log record should trigger an error.

        If the level is greater or equal than the threshold and the message is
        not ignored, create an ErrorInfo object and add it to the BT.

        """
        #error_config = self.agent.error_config_registry

        #if not error_config.should_detect_logged_errors:
        #    return

        levelno = record.levelno
        if levelno < 40:
            return

        msg = record.getMessage()
        #if any(conditions.match(msg, match_cond) for match_cond in error_config.ignored_messages):
        #    return

        if record.exc_info:
            msg += '\n' + logging.Formatter().formatException(record.exc_info)

        display_name = self.LOGGING_ERROR_DISPLAY_NAME.format(logger_name=record.name, level_name=record.levelname)
        #self.bt.add_logged_error(ErrorInfo(msg, display_name, levelno))
        self.agent.logger.info('Logging error added in logs as message {0} display name as {1} and level of error is (2}'.format(msg,display_name,levelno))


def intercept_logging(agent, mod):
    LoggingInterceptor(agent, mod.Logger).attach('callHandlers')







"""Interceptor for httplib/http.client.

"""

from __future__ import unicode_literals
#from agent.internal.proxy import *
from . import HTTPConnectionInterceptor
#import agent


class HttplibConnectionInterceptor(HTTPConnectionInterceptor):
    

    def _putrequest(self, putrequest, connection, method, url, *args, **kwargs):
        #print("INSIDE PUT REQUEST")
        self.agent.get_transaction_context().http_request_type = method  # GET, PUT, POST or DELETE
        exit_call = None
        #self.agent.logger.info('Modulename: HttplibConnectionInterceptor class ')
        #self.agent.logger.info("url value is {0}".format(url))
        with self.log_exceptions():
            bt = self.bt

            #self.agent.logger.info("Modulename: HttplibConnectionInterceptor class inside _putrequest bt value is :{0}".format(bt))
            #self.agent.logger.info("Modulename: HttplibConnectionInterceptor class inside _putrequest method  is :{0}".format(method))
            #self.agent.logger.info("Modulename: HttplibConnectionInterceptor class inside _putrequest connection is :{0}".format(connection))
            #self.agent.logger.info("Modulename: HttplibConnectionInterceptor class inside _putrequest putrequest is :{0}".format(putrequest))
            exit_call = self.http_call_begin(bt, connection.host, url)
            #self.agent.logger.info("Modulename: HttplibConnectionInterceptor class inside _putrequest exit_call is :{0}".format(exit_call))

            if bt:
                scheme = 'https' if self._request_is_https(connection) else 'http'
                connection._pythonagent_exit_call = exit_call
                #backend = self.get_backend(connection.host, connection.port, scheme, url)
                #if backend:
                    #exit_call = self.start_exit_call(bt, backend, operation=url)
                    #connection._pythonagent_exit_call = exit_call
        return putrequest(connection, method, url, pythonagent_exit_call=exit_call, *args, **kwargs)

    def _endheaders(self, endheaders, connection, *args, **kwargs):
        #print("INSIDE END HEADERS")

        #self.agent.logger.info('Modulename: HttplibConnectionInterceptor class inside _endheaders ')

        exit_call = getattr(connection, '_pythonagent_exit_call', None)
        #self.agent.logger.info("Modulename: HttplibConnectionInterceptor class inside _endheaders exit_call is :{0}".format(exit_call))

        with self.log_exceptions():
            header = self.make_correlation_header(exit_call)
            if header is not None:
                connection.putheader(*header)
                #self.agent.logger.debug('Added correlation header to HTTP request: %s, %s' % header)
        return endheaders(connection, pythonagent_exit_call=exit_call, *args, **kwargs)

    def _getresponse(self, getresponse, connection, *args, **kwargs):
        #print("INSIDE GET RESPONSE")
        # CORE-40945 Catch TypeError as a special case for Python 2.6 and call getresponse with just the
        # HTTPConnection instance.
        #self.agent.logger.info('Modulename: HttplibConnectionInterceptor class inside _getresponse')

        exit_call = getattr(connection, '_pythonagent_exit_call', None)
        #self.agent.logger.info("Modulename: HttplibConnectionInterceptor class inside _getresponse exit_call is :{0}".format(exit_call))

        #self.agent.logger.info("exit_call value after getattr,".format(exit_call))
        try:
            #print("inside try")
            with self.end_exit_call_and_reraise_on_exception(exit_call,
                                                             ignored_exceptions=(TypeError,)):
                #self.agent.logger.debug("inside with")
                response = getresponse(connection, *args, **kwargs)
                #print("RESPONSE STATUS CODE: ", response.status)
        except TypeError:
            #print("inside except")
            with self.end_exit_call_and_reraise_on_exception(exit_call):
                #self.agent.logger.debug("inside with")
                response = getresponse(connection)
                #print("RESPONSE STATUS CODE: ", response.status)

        self.http_call_end(self.bt, exit_call, response.status)
        #self.proxy.http_call_end(self.agent.bt,exit_call)
        try:
            del connection._pythonagent_exit_call
        except AttributeError:
            pass
        return response


def intercept_httplib(agent, mod):
    #print("intercept_httplib")
    HTTPConnectionInterceptor.https_connection_classes.add(mod.HTTPSConnection)
    interceptor = HttplibConnectionInterceptor(agent, mod.HTTPConnection)
    interceptor.attach(['putrequest', 'endheaders'])
    interceptor.attach('getresponse', wrapper_func=None)   # CORE-40945 Do not wrap getresponse in the default wrapper.



import json
#import jsonpickle
#from json import JSONEncoder

try:
    from types import SimpleNamespace as Namespace
except ImportError:
    # Python 2.x fallback
    from argparse import Namespace

class MasterObj(object):
    def __init__(self,modules):
        self.modules = modules

class Module(object):
    def __init__(self,moduleLevelMethod,classes,name):
        self.name = name
        self.moduleLevelMethod = moduleLevelMethod
        self.classes = classes
        
class Method(object):
    def __init__(self,name,signature):
        self.name = name
        self.signature = signature

class Class(object):
    def __init__(self,methods,className):
        self.className = className
        self.methods = methods



#m1 = Method("method1","NA")
#m2 = Method("method2","NA")
#classLevelMethods = [m1,m2]
#class1 = Class(classLevelMethods,"classname1")
#classes = [class1]
#moduleLevelMethod = [m1,m2]
#mod1 = Module(moduleLevelMethod,classes,"https")
#modules = [mod1]
#student = MasterObj(modules)

#print("Encode Object into JSON formatted Data using jsonpickle")
def flatten_hook(obj):
    for key, value in obj.items():
        if isinstance(value, str):
            try:
                obj[key] = json.loads(value, object_hook=flatten_hook)
            except ValueError:
                pass
    return obj

#studentJSON = jsonpickle.encode(student)
#print(studentJSON)

#json1 = json.loads(studentJSON,object_hook=flatten_hook)
#print("Decode and Convert JSON into Object using object_hook")
#print(json1)
#print("Object type is: ", type(json1))
#json2 = r"""{
#  "modules": [
#    {
#      "moduleName": "https",
#      "moduleLevelMethod": [
#      {
#        "name":"method1",
#        "signature":"NA"
#      },
#      {
#        "name":"method2",
#        "signature":"NA"
#      }
#      ],
#      "classes": [
#        {
#          "className": "BaseInterceptor",
#          "classLevelMethods": [
#      {
#        "name":"classmethod1",
#        "signature":"NA"
#      },
#      {
#        "name":"classmethod2",
#        "signature":"NA"
#      }
#      ]
#        }
#      ]
#    }
#  ]
#}
# """

def create_method_json(filePath):
    try:
        f = open(filePath, "r")
    except:
        return
    obj2 = json.loads(f.read(),object_hook=flatten_hook)

    return obj2

def create_method_array(filePath):
    #print('reading the profile at path: ',filePath)
    try:
        f = open(filePath, "r")
    except:
        return
    #print('Exception occure to open profiler file')
    obj2 = json.loads(f.read(),object_hook=flatten_hook)

#list of methods
    LOM = []

    for module in obj2["modules"]:
        
        #module = module.split("/")[-1].replace(".py","")
       # print(module+"--------------------------------------------")
        for MLM in module["moduleLevelMethod"]:
            #print(module["moduleName"])
            LOM.append(module["moduleName"]+"."+MLM["name"])
        for each_class in module["classes"]:
            for CLM in each_class["classLevelMethods"]:
                LOM.append(module["moduleName"]+"."+each_class["className"]+"."+CLM["name"])

   # print(LOM)

    arr = [str(r) for r in LOM]

    #print(arr)
    return arr

#abc = create_method_json("/home/cavisson/methodTest/instrumentationprofile.json")["modules"]

#print(abc["moduleName"])

